@@ -1,4 +1,4 @@
-#!/usr/bin/env python3.6
+#!/usr/bin/env python3.7
 
 #  What does this script do?
 # This script takes a pair of read (R1 and R2) in the format of dataframe row
@@ -6,11 +6,11 @@
 
 
 import pandas as pd
-import cigar
 import re
 from Bio.Seq import Seq
 
 from TileSeqMut import posterior
+# import posterior
 
 class MutParser(object):
 
@@ -49,8 +49,10 @@ class MutParser(object):
         # get information for R1
         # where the mapping starts
         self._r1_pos = int(self._reads["pos_start_r1"])
-        self._r1_cigar = list(cigar.Cigar(self._reads["cigar_r1"]).items())
-        self._r1_readlen = sum([i[0] for i in self._r1_cigar])
+        # self._r1_cigar = list(cigar.Cigar(self._reads["cigar_r1"]).items())
+        r1_cigar = [i for i in re.split("([\d]+\S)", self._reads["cigar_r1"]) if i != ""]
+        self._r1_cigar = [(re.split(r"([\d]+)", i)[1], re.split(r"([\d]+)", i)[2]) for i in r1_cigar]
+        self._r1_readlen = sum([int(i[0]) for i in self._r1_cigar])
         self._r1_ref = self._seq[int(self._r1_pos)-1:int(self._r1_pos)+len(self._reads["seq_r1"])]
         self._r1_qual = self._reads["qual_r1"]
         self._r1_read = self._reads["seq_r1"]
@@ -58,8 +60,10 @@ class MutParser(object):
 
         # get the ref sequence for R2
         self._r2_pos = int(self._reads["pos_start_r2"])
-        self._r2_cigar = list(cigar.Cigar(self._reads["cigar_r2"]).items())
-        self._r2_readlen = sum([i[0] for i in self._r2_cigar])
+        # self._r2_cigar = list(cigar.Cigar(self._reads["cigar_r2"]).items())
+        r2_cigar = [i for i in re.split("([\d]+\S)", self._reads["cigar_r2"]) if i != ""]
+        self._r2_cigar = [(re.split(r"([\d]+)", i)[1], re.split(r"([\d]+)", i)[2]) for i in r2_cigar]
+        self._r2_readlen = sum([int(i[0]) for i in self._r2_cigar])
         self._r2_ref = self._seq[int(self._r2_pos)-1:int(self._r2_pos)+len(self._reads["seq_r2"])]
         self._r2_qual = self._reads["qual_r2"]
         self._r2_read = self._reads["seq_r2"]
@@ -67,6 +71,42 @@ class MutParser(object):
 
         return self
 
+    def _convert_mut(self, mut_list, read):
+        """
+        convert a list of mutations into the input format for posterior function
+        return a df with mutations
+        """
+        if mut_list == []:
+            return pd.DataFrame({}, columns=["m", "read", "pos", "ref", "alt", "qual"])
+        final_mut = []
+        for i in mut_list:
+            mut = i.split("|")
+            if "del" in mut:
+                del_pos = int(mut[0])
+                for base in mut[1]:
+                    m = f"{del_pos}|{base}|del|{mut[-1]}"
+                    del_pos+=1
+                    final_mut.append(m)
+            else:
+                final_mut.append(i)
+        final_df = pd.DataFrame({"m":final_mut})
+        final_df["read"] = read
+        final_df[["pos", "ref", "alt", "qual"]] = final_df["m"].str.split("|", expand=True)
+        # validate final_df
+        # for each read, there shouldn't be duplicated positions
+        if final_df.pos.duplicated().any():
+            self._logging.error(f"duplicated positions presented")
+            self._logging.error(final_df)
+            raise DuplicatedMutations()
+
+        # for all the mutations
+        # there shouldn't be any ref == alt
+        same_refalt = final_df[final_df["ref"] == final_df["alt"]]
+        if not same_refalt.empty:
+            raise SameRefandAlt(same_refalt)
+
+        return final_df
+
     def _main(self):
         """
         return a list of mutations from paired reads (R1 and R2)
@@ -75,80 +115,69 @@ class MutParser(object):
         self._get_seq()
 
         # parse mutations in R1
-        r1_snp, r1_delins, r1_map_pos = self._parse_cigar_mdz(self._r1_cigar, self._r1_mdz, self._r1_ref, self._r1_read, self._r1_pos, self._r1_qual)
+        r1_snp, r1_delins, map_pos_r1 = self._parse_cigar_mdz(self._r1_cigar, self._r1_mdz, self._r1_ref, self._r1_read,
+                                                 self._r1_pos,self._r1_qual)
         # parse mutations in R2
-        r2_snp, r2_delins, r2_map_pos = self._parse_cigar_mdz(self._r2_cigar, self._r2_mdz, self._r2_ref, self._r2_read, self._r2_pos, self._r2_qual)
+        r2_snp, r2_delins, map_pos_r2 = self._parse_cigar_mdz(self._r2_cigar, self._r2_mdz, self._r2_ref, self._r2_read,
+                                                  self._r2_pos,self._r2_qual)
 
-        # final list of mutations
-        final_mut = []
-        # for deletion or insertion, we count those that appeared on both reads
-        del_ins = list(set(r1_delins) & set(r2_delins))
-        if r1_snp != [] or r2_snp != []:
-            r1_snp = pd.DataFrame({"snp": r1_snp})
-            r1_snp["read"] = "r1"
-            r2_snp = pd.DataFrame({"snp": r2_snp})
-            r2_snp["read"] = "r2"
-            snp_df = r1_snp.append(r2_snp).reset_index(drop=True)
-            # two cases of snp
-            # mutations are on both reads
-            # mutaiton only shows on one read
+        final_df_r1 = self._convert_mut(r1_snp, "r1")
+        final_df_r2 = self._convert_mut(r2_snp, "r2")
 
-            # d = dict(tuple(df.groupby(df['x'].diff().gt(100).cumsum())))
-            # 1. group mutations by position
-            # split snp column into 3
-            snp_df[["pos", "ref", "alt"]] = snp_df["snp"].str.split("|", expand=True)
-            #print(snp_df)
-            final_mut = []
-            for index, row in snp_df.iterrows():
-                pos = int(row["pos"])
-                if row["alt"] == "N": continue
-                if (r1_map_pos.get(pos) is None) or (r2_map_pos.get(pos) is None):
-                    continue
-                r1_qual = self._r1_qual[r1_map_pos[pos]]
-                r2_qual = self._r2_qual[r2_map_pos[pos]]
-                r1_basecall = self._r1_read[r1_map_pos[pos]]
-                r2_basecall = self._r2_read[r2_map_pos[pos]]
-                if r1_basecall == "N" or r2_basecall == "N": continue
-                wt = row["ref"]
-                pos_prob = posterior.bayesian_variant_call([r1_basecall, r2_basecall], [r1_qual, r2_qual], wt, self._mutrate)
-                #print(pos_prob)
-                # if two mut are different, it will return a dictionary with two keys and their posterior probabilities
-                # pick the one with higher probability
-                # if it is the wt then do nothing
-                # if not, add this mutation to the final list
-                if pos_prob[r1_basecall] > pos_prob[r2_basecall]:
-                    if r1_basecall == wt: continue
-                    if pos_prob[r1_basecall] > self._cutoff:
-                        final_mut.append(row.snp)
-                elif pos_prob[r1_basecall] < pos_prob[r2_basecall]:
-                    if r2_basecall == wt: continue
-                    if pos_prob[r2_basecall] > self._cutoff:
-                        final_mut.append(row.snp)
-                elif pos_prob[r1_basecall] == pos_prob[r2_basecall]:
-                    if r2_basecall == wt or r1_basecall == wt:
-                        print(r1_basecall, r2_basecall, wt)
-                    if pos_prob[r2_basecall] > self._cutoff:
-                        final_mut.append(row.snp)
+        # merge two df
+        snp_df = pd.merge(final_df_r1, final_df_r2, on=["pos"], how="outer", suffixes=('_r1', '_r2'))
+        snp_df["pos"] = pd.to_numeric(snp_df["pos"])
 
-        final_mut += del_ins
-        final_mut = list(set(final_mut))
+        final_df_r1 = self._convert_mut(r1_delins, "r1")
+        final_df_r2 = self._convert_mut(r2_delins, "r2")
+
+        # merge two df
+        delins_df = pd.merge(final_df_r1, final_df_r2, on=["pos"], how="outer", suffixes=('_r1', '_r2'))
+        delins_df["pos"] = pd.to_numeric(delins_df["pos"])
+
+        merged_df = [snp_df, delins_df]
+        merged_df = pd.concat(merged_df)
+        merged_df = merged_df.sort_values(by="pos")
+
+        # group mutations based on positions
+        # any mutations that are within 3bp are grouped together
+        n = 3
+        d = dict(tuple(merged_df.groupby(merged_df['pos'].diff().gt(n).cumsum())))
+        # analyze the dictionary of clusters
+        # and get posterior
+        # print(d, self._mutrate, self._cutoff)
+        pos_df, all_df, clustered_r1, clustered_r2 = posterior.cluster(d, self._r1_qual,self._r2_qual, map_pos_r1,
+                                                                 map_pos_r2, self._mutrate,
+                                           self._cutoff)
+        final_mut = list(set(pos_df.m.tolist()))
         final_mut.sort()
-        hgvs, outside_mut = self._get_hgvs(final_mut)
-        #if "45T1T0A2T0T5A3C0C8A2A0A1A3C4A0C4A2C2A0G2C1A2G4C0T2A5A4" in self._r1_mdz:
-        #  print(final_mut)
-        #  print(snp_df)
-        #  print(r1_delins)
-        #  print(self._r1_cigar)
-        #  print(self._r1_pos)
-        #  print(self._r1_ref)
-        #  print(self._r1_read)
-        #  print(self._r1_mdz)
-        #  print(self._r2_cigar)
-        #  print(self._r2_pos)
-        #  print(self._r2_ref)
-        #  print(self._r2_read)
-        #  print(self._r2_mdz)
-        return hgvs, outside_mut
+
+        final_r1_cluster = list(set(clustered_r1.m.tolist()))
+        final_r1_cluster.sort()
+        final_r2_cluster = list(set(clustered_r2.m.tolist()))
+        final_r2_cluster.sort()
+
+        if final_mut != []:
+            hgvs, outside_mut = self._get_hgvs(final_mut)
+        else:
+            hgvs, outside_mut = [], []
+
+        if final_r1_cluster != []:
+            hgvs_r1_clusters, outside_mut_r1 = self._get_hgvs(final_r1_cluster)
+        else:
+            hgvs_r1_clusters, outside_mut_r1 = [], []
+
+        if final_r2_cluster != []:
+            hgvs_r2_clusters, outside_mut_r2 = self._get_hgvs(final_r1_cluster)
+        else:
+            hgvs_r2_clusters, outside_mut_r2 = [], []
+
+        if "c.463T>G" in hgvs:
+            print(merged_df)
+            # print(pos_df)
+            print(all_df)
+
+        return hgvs, outside_mut, all_df, hgvs_r1_clusters, hgvs_r2_clusters
 
     def _parse_cigar_mdz(self, cigar, mdz_raw, ref, read, pos, qual):
         """
@@ -161,7 +190,7 @@ class MutParser(object):
         # check starting point of the sequence
         # remove soft clipped bases and adjust for position
         if cigar[0][1] == "S": # soft clip occurs in the beginning of a read
-            clip += cigar[0][0]
+            clip += int(cigar[0][0])
 
         # convert mdz string into a list
         mdz = re.findall('.*?[.ATCG]+', mdz_raw)
@@ -203,13 +232,15 @@ class MutParser(object):
             else:
                 # based on number of bases mapped, get the inserted base from read
                 ins_base = read[read_start:read_start+int(i[0])]
+                qual_base = qual[read_start:read_start + int(i[0])]
+                qual_base = ",".join(list(qual_base))
                 # calculate inserted base posterior by using the quality string
                 # qual_ins = qual[mapped:mapped+int(i[0])]
                 # keep the insertion position and inserted lenth in a list
                 ins_pos.append([ins_pos_ref, int(i[0])])
                 # add the insertion to mut_list
                 #mut_list.append([str(pos+mapped+deleted),ins_base,"ins"])
-                delins_list.append(str(pos+mapped+deleted)+"|"+ins_base+"|ins")
+                delins_list.append(f"{pos + mapped + deleted}|{ins_base}|ins|{qual_base}")
                 # skip inserted bases on read
                 read_start += int(i[0])
         # zip two lists into a dictionary
@@ -225,7 +256,7 @@ class MutParser(object):
         inserted_pos = 0
         deleted_len = 0
         map_pos = 0 + clip
-        ##read_ref_map = [] # map [ref_pos, ref_base, read_pos, read_base, read_quality_score]
+
         for i in mdz:
             # for each item in the MD:Z string
             # split the item into number and letter
@@ -236,49 +267,51 @@ class MutParser(object):
             map_pos += match_len# update how many bp are mapped
 
             while ins and map_pos >= ins[0]:
-                #map_pos += ins[1]
                 inserted_pos += ins[1]
                 ins = next(iter_ins, None)
 
             if "^" not in base:
                 read_pos += match_len
                 # this means a single nt change
-                #mut_list.append([str(pos+read_pos-clip),base,read[read_pos+inserted_pos-deleted_len],qual[read_pos+inserted_pos-deleted_len]])
-                snp_list.append(str(pos+read_pos)+"|"+base+"|"+str(read[read_pos+inserted_pos-deleted_len+clip]))
-
-                ## DEBUG ##
-                #if base == str(read[read_pos+inserted_pos-deleted_len+clip]):
-                #     print(snp_list)
-                #     print(delins_list)
-                #     print("map_pos", map_pos)
-                #     print("ins_pos", ins_pos)
-                #     print("ins", ins)
-                #     print("mdz", i)
-                #     print(mdz)
-                #     print(match_len)
-                #     print(read_pos)
-                #     print(inserted_pos)
-                #     print(deleted_len)
-                #     print(read[read_pos+inserted_pos-deleted_len+clip-1], read[read_pos+inserted_pos-deleted_len+clip],read[read_pos+inserted_pos-deleted_len+clip+1])
-                #     ## DEBUG ##
-                #     print(self._r1_ref)
-                #     print(self._r1_read)
-                #     print(self._r1_mdz, self._r1_cigar)
-                #
-                #     print(self._r2_ref)
-                #     print(self._r2_read)
-                #     print(self._r2_mdz, self._r2_cigar)
+                # mut_list.append([str(pos+read_pos-clip),base,read[read_pos+inserted_pos-deleted_len],
+                # qual[read_pos+inserted_pos-deleted_len]])
+                pos_on_read = read_pos+inserted_pos-deleted_len+clip
+                snp_list.append(f"{pos+read_pos}|{base}|{read[pos_on_read]}|{qual[pos_on_read]}")
 
+                # adjust read pos with 1bp change (move to the right for 1 pos)
                 map_pos += len(base)
-                read_pos += 1 # adjust read pos with 1bp change (move to the right for 1 pos)
+                read_pos += 1
             else: # deletion
                 read_pos += match_len
                 #mut_list.append([str(pos+read_pos-clip),base[1:],"del"])
-                delins_list.append(str(pos+read_pos)+"|"+base[1:]+"|"+"del")
+                # quality of the base before deleted base
+                qual_pre = qual[read_pos+inserted_pos-deleted_len+clip-1]
+                # quality of the base after deleted base
+                qual_post = qual[read_pos+inserted_pos-deleted_len+clip]
+                delins_list.append(f"{pos + read_pos}|{base[1:]}|del|{qual_pre},{qual_post}")
 
                 deleted_len += len(base[1:])
                 read_pos += len(base[1:])
                 map_pos += len(base[1:])
+
+        # if "287|T|C|E" in snp_list:
+        #     print(snp_list, delins_list)
+        #     print(self._r1_pos,
+        #     self._r1_cigar,
+        #     self._r1_readlen,
+        #     self._r1_ref,
+        #     self._r1_qual,
+        #     self._r1_read ,
+        #     self._r1_mdz)
+        #
+        #     print(self._r2_pos,
+        #           self._r2_cigar,
+        #           self._r2_readlen,
+        #           self._r2_ref,
+        #           self._r2_qual,
+        #           self._r2_read,
+        #           self._r2_mdz)
+
         return snp_list, delins_list, pos_map
 
     def _get_hgvs(self, mut_list):
@@ -287,11 +320,8 @@ class MutParser(object):
         translate a list of mut to aa changes
         return list of nt changes represent by hgvs strings
         """
-        # there are two types of mutations in mut list
-        # ins del
-
-        # how to track consecutive changes?
-        concecutive_snp = [] # if the snp changes are concecutive, it will be represent as delins
+        # if the snp changes are concecutive, it will be represent as delins
+        concecutive_snp = []
         combined_snp = ""
 
         # track mutation positions that was not within the tiled region
@@ -302,7 +332,9 @@ class MutParser(object):
         # also record any snp that happened in the range
         delins = []
         mutations = []
+
         for mut in mut_list:
+
             mut_change = mut.split("|")
             tmp_pos = int(mut_change[0]) # template position
 
@@ -310,6 +342,7 @@ class MutParser(object):
             try:
                 cds_pos = self._seq_lookup[self._seq_lookup.temp_pos == tmp_pos].cds_pos.values.item()
             except:
+                # this means that the mutation was on the template
                 continue
 
             if cds_pos < self._tile_begins or cds_pos > self._tile_ends:
@@ -322,6 +355,7 @@ class MutParser(object):
 
             if ("del" in mut) or ("ins" in mut): # deletion or insertion
                 mut_change[0] = cds_pos
+
                 if delins == []: # nothing is on track
                     delins.append(mut_change)
                 else: # compare bases
@@ -354,6 +388,7 @@ class MutParser(object):
 
                 ## validate: if the reference base matches the ref in seq_lookup
                 if cds_ref != mut_change[1]:
+                    print(self._reads)
                     raise ValueError(f"Reference base - pos {tmp_pos}, base {mut_change[1]} does not match the reference provided by user - base {cds_ref}")
                 # track concecutive changes
                 if len(concecutive_snp) == 0:
@@ -377,6 +412,7 @@ class MutParser(object):
                         concecutive_snp = [cds_pos]
                         combined_snp = mut_change[2]
                         mutations.append(hgvs)
+
         if len(concecutive_snp) !=0:
             hgvs = snp_to_hgvs(concecutive_snp, combined_snp, self._cds)
             mutations.append(hgvs)
@@ -390,18 +426,29 @@ class MutParser(object):
         else:
             joined = ";".join(mutations)
             mutations = f"c.[{joined}]"
+
+        # output_df = pd.DataFrame(output_df, columns=["mut", "hgvs"])
+        # print(output_df)
+
         return mutations, outside_mut
 
+
 def snp_to_hgvs(concec_pos, combined_bases, cds):
     """
     helper function to obtain hgvs sstring given a list of positions and a string of bases combined.
 
+    @param concec_pos
+    @param combined_bases
+    @param cds
+
+    @return
     """
     if len(concec_pos) == 1:
         ref_base = cds[concec_pos[0]-1]
         hgvs = f"{concec_pos[0]}{ref_base}>{combined_bases}"
     else:
         hgvs = f"{concec_pos[0]}_{concec_pos[-1]}delins{combined_bases}"
+
     return hgvs
 
 
@@ -413,7 +460,7 @@ def delins_to_hgvs(cds_seq, delins):
     # sort delins
     delins.sort()
     # convert delins to df
-    delins_df = pd.DataFrame(delins, columns=["pos", "base", "type"])
+    delins_df = pd.DataFrame(delins, columns=["pos", "base", "type", "qual"])
     types = delins_df["type"].unique()
 
     if len(types) == 1:
@@ -459,6 +506,34 @@ def delins_to_hgvs(cds_seq, delins):
         #print("del+ins, ",hgvs)
     return hgvs
 
+
+class DuplicatedMutations(Exception):
+    """Exception raised when there are two mutations found on the same read at the same positions.
+
+        Attributes:
+            message -- explanation of the error
+        """
+
+    def __init__(self, message="Same position mutated twice in one read!"):
+        self.message = message
+        super().__init__(self.message)
+
+
+class SameRefandAlt(Exception):
+    """Exception raised when reference base and alt base are the same.
+
+        Attributes:
+            message -- explanation of the error
+        """
+
+    def __init__(self, mut_df, message="Same reference base and alt base!"):
+        self.mut = mut_df
+        self.message = message
+        super().__init__(self.message)
+
+    def __str__(self):
+        return f'{self.mut} -> {self.message}'
+
 if __name__ == "__main__":
 
     # test delins_to_hgvs
